/**********************************************************************
  VTT_voice_to_text_7.ino
  Code by lingib
  https://www.instructables.com/member/lingib/instructables/
  Last update 24 May 2020

  ------
  About
  ------
  VTT_voice_to_text_7.ino is a Bluetooth speech interpreter used for 
  controlling each of five stepping motors attached to a robot hand.

  The speech library contains a number of keywords that trigger some flags. 
  The Arduino interpreter determines what action is required by looking at these flags.
 
   ----------
  COPYRIGHT
  ----------
  This code is free software: you can redistribute it and/or
  modify it under the terms of the GNU General Public License as published
  by the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This software is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License. If
  not, see <http://www.gnu.org/licenses/>.
***// -------------------
// Globals
// -------------------

// ----- bit set/clear/check/toggle macros
#define SET(x,y) (x |=(1<<y))
#define CLR(x,y) (x &= (~(1<<y)))
#define CHK(x,y) (x & (1<<y))
#define TOG(x,y) (x^=(1<<y))

// ----- Bluetooth
#include <SoftwareSerial.h>                                 // this frees normal port for Debugging
const byte TXD = 13;                                        // connect HC-06 TXD pin to Arduino pin 13
const byte RXD = 12;                                        // connect HC-06 RXD pin to Arduino pin 12 via resistive divider
SoftwareSerial Bluetooth(TXD, RXD);                         // format: bluetooth-name(HC-06 TXD pin, HC-06 RXD pin)

// ----- finger connections to Big Easydriver motor controllers
const byte Thumb_step = 11;                                 // Thumb
const byte Thumb_dir = 10;
const byte Index_step = 9;                                  // index finger
const byte Index_dir = 8;
const byte Middle_step = 7;                                 // middle finger
const byte Middle_dir = 6;
const byte Ring_step = 5;                                   // ring finger
const byte Ring_dir = 4;
const byte Little_step = 3;                                 // little finger
const byte Little_dir = 2;

// ----- motor definitions
const long Steps_per_mm = 200 * 16 / 40L;                   // 200steps/rev; 16 x microstepping; 40mm/rev
const unsigned long Pulse_width = 2;                        // easydriver step pulse-width (uS)
const unsigned long Delay = 300;                            // delay (uS) between motor steps
const bool CW = true;                                       // motors (looking at shaft) turn CW when DIR pin is high
const bool CCW = !CW;
const short Default_steps = 2500;
short Steps_required = Default_steps;                       // number of steps to open/close the fingers
short Pick_index_steps = 2800;                              // number of steps to assume pink position
short Pick_thumb_steps = 1750;
bool Pick_flag = false;

// ----- motor control
bool THUMB_step = false;
bool THUMB_dir = CCW;
bool INDEX_step = false;
bool INDEX_dir = CCW;
bool MIDDLE_step = false;
bool MIDDLE_dir = CCW;
bool RING_step = false;
bool RING_dir = CCW;
bool LITTLE_step = false;
bool LITTLE_dir = CCW;

// ----- interpreter definitions
const short String_size = 128;                              // string size
char Buffer[String_size + 1];                               // Buffer
short Buffer_index = 0;                                     // Buffer index
char Input_char;                                            // bluetooth character
String Input_string;                                        // bluetooth string

// ----- flags
/* the keywords are divided into groups */
const String Keywords[] =  {
  "thumb", "index", "middle", "ring", "little",           // specific finger(s)
  "hand", "fingers",  "release",                          // all fingers
  "open", "raise", "extend",                              // open
  "close", "lower",                                       // close
  "carry", "hold", "pick", "demo", "calibrate",           // tasks
};
bool Keyword_flag[] = {                                   // flag array for keywords
  0, 0, 0, 0, 0,                                          // "thumb", "index", "middle", "ring", "little"
  0, 0, 0,                                                // "hand", "fingers", "release"
  0, 0, 0,                                                // "open", "raise", "extend"
  0, 0,                                                   // "close", "lower"
  0, 0, 0, 0, 0                                           // "carry", "hold", "pick", "demo", "calibrate"
};

bool Open_flag = false;                                   // SET if the words "open", "raise", "extend", release are spoken. */
bool Close_flag = false;                                  // SET if the words "close", "lower", "bend", are spoken.
bool Hand_flag = false;                                   // SET if the words "hand", "fingers" are spoken.
bool Finger_flag = false;                                 // SET if the words "thumb", "index", "middle", "ring", "little" are spoken */
bool Task_flag = false;                                   // SET if the words "carry", "hold", "pick", "demo", "calibrate" are spoken

bool Task_complete = false;                              // SET if command recognized
bool Fingers_to_move[] = {0, 0, 0, 0, 0};                 // 1 = move finger; 0 = don't move finger
bool Finger_status[] = {1, 1, 1, 1, 1};                   // thumb, index, middle, ring, and little fingers are extended at startup
const bool Open = true;                                   // finger open
const bool Closed = !Open;                                // finger closed

// ----- Debug
bool Debug = true;                                          // "Debug" flag

// -------------------
// setup()
// -------------------

void setup()
{
  // ----- Serial
  Serial.begin(9600);                                       // used for Debugging

  // ----- Bluetooth
  pinMode(TXD, INPUT);                                      // HC-06 send  ...  Arduino receive
  pinMode(RXD, OUTPUT);                                     // HC-06 receive ...Arduino transmit
  Bluetooth.begin(9600);                                    // start bluetooth

  // ----- initialise thumb motor
  pinMode(Thumb_dir, OUTPUT);
  pinMode(Thumb_step, OUTPUT);
  digitalWrite(Thumb_dir, CCW);
  delayMicroseconds(Pulse_width);
  digitalWrite(Thumb_step, LOW);

  // ----- initialise index finger motor
  pinMode(Index_dir, OUTPUT);
  pinMode(Index_step, OUTPUT);
  digitalWrite(Index_dir, CCW);
  delayMicroseconds(Pulse_width);
  digitalWrite(Index_step, LOW);

  // ----- initialise middle finger motor
  pinMode(Middle_dir, OUTPUT);
  pinMode(Middle_step, OUTPUT);
  digitalWrite(Middle_dir, CCW);
  delayMicroseconds(Pulse_width);
  digitalWrite(Middle_step, LOW);

  // ----- initialise ring finger motor
  pinMode(Ring_dir, OUTPUT);
  pinMode(Ring_step, OUTPUT);
  digitalWrite(Ring_dir, CCW);
  delayMicroseconds(Pulse_width);
  digitalWrite(Ring_step, LOW);

  // ----- initialise little finger motor
  pinMode(Little_dir, OUTPUT);
  pinMode(Little_step, OUTPUT);
  digitalWrite(Little_dir, CCW);
  delayMicroseconds(Pulse_width);
  digitalWrite(Little_step, LOW);
}

// -------------------
// main loop()
// -------------------

void loop()
{
  // Serial.println(F("test"));
  //  delay(1000);
  // ----- wait for Bluetooth data
  while (Bluetooth.available())
  {
    Input_char = (char)Bluetooth.read();                    // read character
    Serial.write(Input_char);                               // echo character to the screen
    Buffer[Buffer_index++] = Input_char;                           // add char to Buffer
    if (Input_char == '\n')                                 // check for line feed
    {
      Input_string = Buffer;                                // convert to command string
      process(Input_string);                                // perform command
      Input_string = "";                                    // clear command string
      memset(Buffer, '\0', sizeof(Buffer));                 // clear Buffer
      Buffer_index = 0;                                     // point to Buffer start
    }
  }
}

//--------------------------------------------------------------------------
// process()
//--------------------------------------------------------------------------
void process(String string)
{
  /* This routine expects the thumb and fingers to be extended at start-up */

  if (Debug) Serial.println(string);

  if (Pick_flag) {

    ////////////////////////////
    // pick exit-routine
    ////////////////////////////

    /* This routine is required because the pick command uses custom cable lengths */

    // ----- clear flags
    Open_flag = false;
    Close_flag = false;
    Hand_flag = false;
    Finger_flag = false;
    Task_flag = false;
    Task_complete = false;
    Fingers_to_move[0] = false;
    Fingers_to_move[1] = false;
    Fingers_to_move[2] = false;
    Fingers_to_move[3] = false;
    Fingers_to_move[4] = false;

    // ----- open index finger
    Open_flag = true;
    Fingers_to_move[1] = true;                    // open index finger
    Steps_required = Pick_index_steps;
    move_fingers();

    // ----- now open the thumb
    Fingers_to_move[0] = true;                    // open thumb
    Fingers_to_move[1] = false;
    Steps_required = Pick_thumb_steps;
    move_fingers();

    // ----- house-keeping
    Open_flag = false;
    for (short i = 0; i < 5; i++) {
      Finger_status[i] = true;                    // all fingers open
    }
    for (short i = 0; i < 5; i++) {
      Fingers_to_move[i] = false;                 // no fingers to move
    }
    Steps_required = Default_steps;               // restore default steps
    Pick_flag = false;                            // clear flag

  } else {

    ////////////////////////////
    // default routine
    ////////////////////////////

    // ----- locals
    short string_length;
    short keyword_length;
    short substring_length;
    short number_keywords;

    // ----- preprocess input string
    string.toLowerCase();                                           // convert string to lowercase
    string.trim();                                                  // remove line-feed
    string_length = string.length();                                // calculate actual number of characters
    number_keywords = sizeof(Keywords) / sizeof(Keywords[0]);       // calculate number of keywords ... this works because Keywords[] is an array of fixed length pointers

    // ----- zero flags
    Open_flag = false;                            // SET if the words "open", "raise", "extend" are spoken. */
    Close_flag = false;                           // SET if the words "close", "lower", "bend", are spoken.
    Hand_flag = false;                            // SET if the words "hand", "fingers" are spoken.
    Finger_flag = false;                          // SET if the words "thumb", "index", "middle", "ring", "little" are spoken */
    Task_flag = false;                            // SET if the words "carry", "hold", "pick", "demo", "calibrate" are spoken
    Task_complete = false;                       // SET if command recognized

    for (int i = 0; i < number_keywords; i++) {   // SET if keyword detected
      Keyword_flag[i] = false;
    }

    for (short i = 0; i < 5; i++) {
      Fingers_to_move[i] = false;                 // SET coresponding flag if "thumb", "index", "middle", "ring", "little" finger positions are to change
    }

    ///////////////////////////
    // Search for keywords
    ///////////////////////////

    // ----- search the input string for keywords
    for (short i = 0; i < number_keywords; i++) {
      keyword_length = Keywords[i].length();                                // calculate keyword length
      if (keyword_length <= string_length) {                                // validate search
        for (short j = 0; j < string_length - keyword_length + 1; j++) {    // start search
          if (Keywords[i] == string.substring(j, j + keyword_length)) {     // look for pattern match
            Keyword_flag[i] = true;                                         // set keyword flag
          }
        }
      }
    }

    ///////////////////////////
    // Set flags
    ///////////////////////////

    // ----- open flag
    /* SET if the words "release", "open", "raise", "extend" are spoken. */
    if (Keyword_flag[7] ||  Keyword_flag[8] || Keyword_flag[9] || Keyword_flag[10]) Open_flag = true;

    // ----- close flag
    /* SET if the words "close", "lower" are spoken. */
    if (Keyword_flag[11] ||  Keyword_flag[12]) Close_flag = true;

    // ----- hand flag
    /* SET if the words "hand", "fingers", "release", "open", "close" are spoken. */
    if (Keyword_flag[5] ||  Keyword_flag[6] ||  Keyword_flag[7] ||  Keyword_flag[8] ||  Keyword_flag[11]) Hand_flag = true;

    // ----- finger flag
    /* SET if the words "thumb", "index", "middle", "ring", "little" are spoken */
    if (Keyword_flag[0] || Keyword_flag[1] || Keyword_flag[2] || Keyword_flag[3] || Keyword_flag[4]) Finger_flag = true;

    // ----- task flag
    /* SET if the words "carry", "hold", "pick", "demo", "calibrate" are spoken */
    if (Keyword_flag[13] || Keyword_flag[14] || Keyword_flag[15] || Keyword_flag[16] || Keyword_flag[17]) Task_flag = true;


    /////////////////////////
    // tasks
    /////////////////////////

    if (Task_flag) {
      // ----- carry
      if (Keyword_flag[13]) {
        carry();
        Task_complete = true;
      }
      // ----- hold
      if (Keyword_flag[14]) {
        Serial.println(F("hold mode"));
        hold();
        Task_complete = true;
      }
      // ----- pick / unpick
      if (Keyword_flag[15]) {
        Serial.println(F("pick mode"));
        pick();
        Task_complete = true;
      }
      // ----- demo
      if (Keyword_flag[16]) {
        Serial.println(F("demo mode"));
        demo();
        Task_complete = true;
      }
      // ----- calibrate
      if (Keyword_flag[17]) {
        Serial.println("calibration mode");
        calibrate();
        Task_complete = true;
      }
    }

    ///////////////////////////////
    // finger commands
    ///////////////////////////////

    if (Finger_flag) {
      Hand_flag = false;                                  // required in case you have used the word "fingers"

      //////////////////////////
      // open specific fingers
      //////////////////////////

      if (Open_flag) {
        // ----- thumb
        if (Keyword_flag[0]) {
          if (Finger_status[0]) {
            Serial.println(F("my thumb is already open"));
          } else {
            Serial.println(F("opening my thumb"));
            Fingers_to_move[0] = true;
            Finger_status[0] = Open;
          }
        }
        // ----- index
        if (Keyword_flag[1]) {
          if (Finger_status[1]) {
            Serial.println(F("my index finger is already open"));
          } else {
            Serial.println(F("opening my index finger"));
            Fingers_to_move[1] = true;
            Finger_status[1] = Open;
          }
        }
        // ----- middle
        if (Keyword_flag[2]) {
          if (Finger_status[2]) {
            Serial.println(F("my middle finger is already open"));
          } else {
            Serial.println(F("opening my middle finger"));
            Fingers_to_move[2] = true;
            Finger_status[2] = Open;
          }
        }
        // ----- ring
        if (Keyword_flag[3]) {
          if (Finger_status[3]) {
            Serial.println(F("my ring finger is already open"));
          } else {
            Serial.println(F("opening my ring finger"));
            Fingers_to_move[3] = true;
            Finger_status[3] = Open;
          }
        }
        // ----- little
        if (Keyword_flag[4]) {
          if (Finger_status[4]) {
            Serial.println(F("my little finger is already open"));
          } else {
            Serial.println(F("opening my little finger"));
            Fingers_to_move[4] = true;
            Finger_status[4] = Open;
          }
        }
      }

      ///////////////////////////
      // close specific fingers
      ///////////////////////////

      if (Close_flag) {
        // ----- thumb
        if (Keyword_flag[0]) {
          if (!Finger_status[0]) {
            Serial.println(F("my thumb is already closed"));
          } else {
            Serial.println(F("closing my thumb"));
            Fingers_to_move[0] = true;
            Finger_status[0] = Closed;
          }
        }
        // ----- index
        if (Keyword_flag[1]) {
          if (!Finger_status[1]) {
            Serial.println(F("my index finger is already closed"));
          } else {
            Serial.println(F("closing my index finger"));
            Fingers_to_move[1] = true;
            Finger_status[1] = Closed;
          }
        }
        // ----- middle
        if (Keyword_flag[2]) {
          if (!Finger_status[2]) {
            Serial.println(F("my middle finger is already closed"));
          } else {
            Serial.println(F("closing my middle finger"));
            Fingers_to_move[2] = true;
            Finger_status[2] = Closed;
          }
        }
        // ----- ring
        if (Keyword_flag[3]) {
          if (!Finger_status[3]) {
            Serial.println(F("my ring finger is already closed"));
          } else {
            Serial.println(F("closing my ring finger"));
            Fingers_to_move[3] = true;
            Finger_status[3] = Closed;
          }
        }
        // ----- little
        if (Keyword_flag[4]) {
          if (!Finger_status[4]) {
            Serial.println(F("my little finger is already closed"));
          } else {
            Serial.println(F("closing my little finger"));
            Fingers_to_move[4] = true;
            Finger_status[4] = Closed;
          }
        }
      }
      move_fingers();
      Task_complete = true;
    }

    /////////////////////////
    // hand commands
    /////////////////////////

    if (Hand_flag) {

      ////////////////////////////
      // open any closed fingers
      ////////////////////////////

      if (Open_flag)  {
        // ----- is hand already open
        if ( (Finger_status[0] == Open) && (Finger_status[1] == Open) && (Finger_status[2] == Open) && (Finger_status[3] == Open) && (Finger_status[4] == Open)) {
          Serial.println(F("my hand is already open"));
        } else {
          // -----
          if (Finger_status[0] == Closed) {
            Serial.println(F("opening my thumb"));
            Fingers_to_move[0] = true;
            Finger_status[0] = Open;
          }
          if (Finger_status[1] == Closed) {
            Serial.println(F("opening my index finger"));
            Fingers_to_move[1] = true;
            Finger_status[1] = Open;
          }
          if (Finger_status[2] == Closed) {
            Serial.println(F("opening my middle finger"));
            Fingers_to_move[2] = true;
            Finger_status[2] = Open;
          }
          if (Finger_status[3] == Closed) {
            Serial.println(F("opening my ring finger"));
            Fingers_to_move[3] = true;
            Finger_status[3] = Open;
          }
          if (Finger_status[4] == Closed) {
            Serial.println(F("opening my little finger"));
            Fingers_to_move[4] = true;
            Finger_status[4] = Open;
          }
          Serial.println(F("my hand is now open"));
        }
      }

      ///////////////////////////////
      // close hand command
      ///////////////////////////////

      if (Close_flag) {
        if ( (Finger_status[0] == Closed) && (Finger_status[1] == Closed) && (Finger_status[2] == Closed) && (Finger_status[3] == Closed) && (Finger_status[4] == Closed)) {
          Serial.println(F("my hand is already closed"));
        } else {
          // ----- thumb
          if (Finger_status[0] == Open) {
            Serial.println(F("closing my thumb"));
            Fingers_to_move[0] = true;
            Finger_status[0] = Closed;
          }
          // ----- index finger
          if (Finger_status[1] == Open) {
            Serial.println(F("closing my index finger"));
            Fingers_to_move[1] = true;
            Finger_status[1] = Closed;
          }
          // ----- middle finger
          if (Finger_status[2] == Open) {
            Serial.println(F("closing my middle finger"));
            Fingers_to_move[2] = true;
            Finger_status[2] = Closed;
          }
          // ------ ring finger
          if (Finger_status[3] == Open) {
            Serial.println(F("closing my ring finger"));
            Fingers_to_move[3] = true;
            Finger_status[3] = Closed;
          }
          // ----- little finger
          if (Finger_status[4] == Open) {
            Serial.println(F("closing my little finger"));
            Fingers_to_move[4] = true;
            Finger_status[4] = Closed;
          }
          Serial.println(F("my hand is now closed"));
        }
      }
      move_fingers();
      Task_complete = true;
    }

    /////////////////////////
    // invalid commands
    /////////////////////////

    if (!Task_complete) {
      Serial.println(F("I do not understand your command"));
    }

    if (Debug) {
      Serial.println(F(""));
      // ----- set keyword column-width to 14 characters
      for (short i = 0; i < number_keywords; i++) {
        short number_chars = Keywords[i].length();
        for (short j = 0; j < 14 - number_chars; j++) {
          Serial.print(" ");                                          // pad each keyword with blank spaces
        }
        Serial.print(Keywords[i]); Serial.print(F(": ")); Serial.println(Keyword_flag[i]);
      }
      Serial.println(F(""));

      Serial.print(F("     open flag: ")); Serial.println(Open_flag);
      Serial.print(F("    close flag: ")); Serial.println(Close_flag);
      Serial.print(F("     hand flag: ")); Serial.println(Hand_flag);
      Serial.print(F("   finger flag: ")); Serial.println(Finger_flag);
      Serial.print(F("     task flag: ")); Serial.println(Task_flag);
      Serial.println(F(""));

      Serial.print(F("  thumb status: ")); Serial.println(Finger_status[0]);
      Serial.print(F("  index status: ")); Serial.println(Finger_status[1]);
      Serial.print(F(" middle status: ")); Serial.println(Finger_status[2]);
      Serial.print(F("   ring status: ")); Serial.println(Finger_status[3]);
      Serial.print(F(" little status: ")); Serial.println(Finger_status[4]);
      Serial.println(F(""));

      Serial.print(F("    move thumb: ")); Serial.println(Fingers_to_move[0]);
      Serial.print(F("    move index: ")); Serial.println(Fingers_to_move[1]);
      Serial.print(F("   move middle: ")); Serial.println(Fingers_to_move[2]);
      Serial.print(F("     move ring: ")); Serial.println(Fingers_to_move[3]);
      Serial.print(F("   move little: ")); Serial.println(Fingers_to_move[4]);
      Serial.println(F(""));
    }

    Serial.println(F(""));
  }
}

// ----------------------------------
// step_motors()
// ----------------------------------
void step_motors() {

  // ----- locals
  enum {index_dir, index_step, thumb_dir, thumb_step};                                      // define bit positions PB0, PB1, PB2, PB3, PB4, PB5
  enum {rx, tx, little_dir, little_step, ring_dir, ring_step, middle_dir, middle_step,};    // define bit positions PD0, PD1, PD2, PD3, PD4, PD5, PD6, PD7
  byte pattern1 = PORTB;                                                                    // read current state PORTB
  byte pattern2 = PORTD;                                                                    // read current state PORTD

  // ----- set motor directions
  /* a direction must be set ... even if it is just a dummy direction */
  (THUMB_dir == CCW)  ? SET(pattern1, thumb_dir)  : CLR(pattern1, thumb_dir);                // fingers close when DIR pins are set to CCW
  (INDEX_dir == CCW)  ? SET(pattern1, index_dir)  : CLR(pattern1, index_dir);
  (MIDDLE_dir == CCW) ? SET(pattern2, middle_dir) : CLR(pattern2, middle_dir);
  (RING_dir == CCW)   ? SET(pattern2, ring_dir)   : CLR(pattern2, ring_dir);
  (LITTLE_dir == CCW) ? SET(pattern2, little_dir) : CLR(pattern2, little_dir);
  PORTB = pattern1;
  PORTD = pattern2;
  delayMicroseconds(Pulse_width);                                                           // wait 2uS for direction lines to stabilise

  // ----- create leading edge of step pulse(s)
  /* only step the motors associated with a moving finger */
  (THUMB_step == true)  ? SET(pattern1, thumb_step)   : CLR(pattern1, thumb_step);          // prepare step pulses
  (INDEX_step == true)  ? SET(pattern1, index_step)   : CLR(pattern1, index_step);
  (MIDDLE_step == true) ? SET(pattern2, middle_step)  : CLR(pattern2, middle_step);
  (RING_step == true)   ? SET(pattern2, ring_step)    : CLR(pattern2, ring_step);
  (LITTLE_step == true) ? SET(pattern2, little_step)  : CLR(pattern2, little_step);
  PORTB = pattern1;                                                                         // step the motors
  PORTD = pattern2;
  delayMicroseconds(Pulse_width);                                                           // mandatory delay

  // ----- create trailing-edge of step-pulse(s)
  CLR(pattern1, thumb_step);                                                                // prepare step lines
  CLR(pattern1, index_step);
  CLR(pattern2, middle_step);
  CLR(pattern2, ring_step);
  CLR(pattern2, little_step);
  PORTB = pattern1;                                                                         // lower step lines
  PORTD = pattern2;

  // ----- determines plotting speed
  delayMicroseconds(Delay);
}

// --------------------------------
// move_fingers()
// --------------------------------
void move_fingers() {
  // ----- locals
  enum {index_dir, index_step, thumb_dir, thumb_step};                                      // define bit positions PB0, PB1, PB2, PB3, PB4, PB5
  enum {rx, tx, little_dir, little_step, ring_dir, ring_step, middle_dir, middle_step,};    // define bit positions PD0, PD1, PD2, PD3, PD4, PD5, PD6, PD7
  byte pattern1 = PORTB;                                                                    // read current state PORTB
  byte pattern2 = PORTD;                                                                    // read current state PORTD

  for (short i = 0; i < Steps_required; i++) {


    // ----- set motor directions
    if (Close_flag) {
      SET(pattern1, thumb_dir);
      SET(pattern1, index_dir);
      SET(pattern2, middle_dir);
      SET(pattern2, ring_dir);
      SET(pattern2, little_dir);
    }
    if (Open_flag) {
      CLR(pattern1, thumb_dir);
      CLR(pattern1, index_dir);
      CLR(pattern2, middle_dir);
      CLR(pattern2, ring_dir);
      CLR(pattern2, little_dir);
    }
    PORTB = pattern1;
    PORTD = pattern2;
    delayMicroseconds(Pulse_width);                                                           // wait 2uS for direction lines to stabilise

    // ----- create leading edge of step pulse(s)
    /* only step the motors associated with a moving finger */
    (Fingers_to_move[0]) ? SET(pattern1, thumb_step)   : CLR(pattern1, thumb_step);          // prepare step pulses
    (Fingers_to_move[1]) ? SET(pattern1, index_step)   : CLR(pattern1, index_step);
    (Fingers_to_move[2]) ? SET(pattern2, middle_step)  : CLR(pattern2, middle_step);
    (Fingers_to_move[3]) ? SET(pattern2, ring_step)    : CLR(pattern2, ring_step);
    (Fingers_to_move[4]) ? SET(pattern2, little_step)  : CLR(pattern2, little_step);
    PORTB = pattern1;                                                                         // step the motors
    PORTD = pattern2;
    delayMicroseconds(Pulse_width);                                                           // mandatory delay

    // ----- create trailing-edge of step-pulse(s)
    CLR(pattern1, thumb_step);                                                                // prepare step lines
    CLR(pattern1, index_step);
    CLR(pattern2, middle_step);
    CLR(pattern2, ring_step);
    CLR(pattern2, little_step);
    PORTB = pattern1;                                                                         // lower step lines
    PORTD = pattern2;

    // ----- determines plotting speed
    delayMicroseconds(Delay);
  }

  // ----- clear Fingers_to_move[] flags
  for (short i = 0; i < 5; i++) {                                                             // needed for recursion
    Fingers_to_move[i] = false;
  }
}

// ----------------
// carry()
// ----------------
void carry() {
  process(F("open your hand"));                                                                  // open any fingers that may be closed
  process(F("close your index middle ring and little fingers"));
}

// ----------------
// hold()
// ----------------
void hold() {
  process(F("open your hand"));                                                                  // open any fingers that may be closed
  process(F("close your thumb and little finger"));                                              // prepare to hold
  process(F("close your index middle and ring fingers"));                                        // close the remaining fingers
}

// -------------------------------
// pick()
// -------------------------------
/* This command requires custom cable lengths */
void pick() {

  // ----- pick
  /* move thumb and indexfinger to pick position */
  process(F("open your fingers"));                                                             // open any closed fingers
  Steps_required = Pick_thumb_steps;                                                           // move thumb to pick position
  process(F("close your thumb"));                                                              // open any fingers that may be closed
  Steps_required = Pick_index_steps;                                                           // move thumb to pick position
  process(F("close your index finger"));
  Steps_required = Default_steps;                                                              // restore Steps_required
  Pick_flag = true;

  Task_complete = true;
}

// -------------------------------
// demo()
// -------------------------------
void demo() {
  process(F("close  thumb"));
  process(F("open thumb"));

  process(F("close index"));
  process(F("open index"));

  process(F("close middle"));
  process(F("open middle"));

  process(F("close ring"));
  process(F("open ring"));

  process(F("close little"));
  process(F("open little"));
}

// ----------------------------------
// calibrate()
// ----------------------------------
void calibrate()
/*
    This routine allows you to determine the number of steps
    that each motor must make to achieve the desired hand-position
*/
{
  // ----- instructions
  Serial.println(F("            MENU"));
  Serial.println(F("  ---------------------------"));
  Serial.println(F("       Up   Down   Finger"));
  Serial.println(F("  ---------------------------"));
  Serial.println(F("   <- '1'  'Q' ->  Thumb"));
  Serial.println(F("   <- '2'  'W' ->  Index"));
  Serial.println(F("   <- '3'  'E' ->  Middle"));
  Serial.println(F("   <- '4'  'R' ->  Ring"));
  Serial.println(F("   <- '5'  'T' ->  Little"));
  Serial.println(F("   Press 'X' to eXit"));
  Serial.println(F(" "));
  Serial.println(F("  Multiple key presses may be"));
  Serial.println(F("  obtained by holding the key"));
  Serial.println(F("  down before pressing <ENTER>"));
  Serial.println(F(" "));

  // ----- flush the Buffer
  while (Serial.available() > 0) Serial.read();

  // ----- control motors with numeric '1'..'5' and alpha 'QWERT' keys
  char keystroke = ' ';
  while (keystroke != 'X') {            //press 'X' key to eXit

    // ----- check for keypress
    if (Serial.available() > 0) {
      keystroke = (char) Serial.read();
    }

    // ----- select task
    switch (keystroke) {
      case '!':
      case '1': {
          // ----- Thumb Up
          Serial.println(F("Thumb up"));
          thumb_up();
          keystroke = ' ';
          break;
        }
      case 'Q':
      case 'q': {
          // ----- Thumb Down
          Serial.println(F("Thumb down"));
          thumb_down();
          keystroke = ' ';
          break;
        }
      case '@':
      case '2': {
          // ----- Index Up
          Serial.println(F("Index up"));
          index_up();
          keystroke = ' ';
          break;
        }
      case 'W':
      case 'w': {
          // ----- Index Down
          Serial.println(F("Index down"));
          index_down();
          keystroke = ' ';
          break;
        }
      case '#':
      case '3': {
          // ----- Middle Up
          Serial.println(F("Middle up"));
          middle_up();
          keystroke = ' ';
          break;
        }
      case 'E':
      case 'e': {
          // ----- Middle Down
          Serial.println(F("Middle down"));
          middle_down();
          keystroke = ' ';
          break;
        }
      case '$':
      case '4': {
          // ----- Ring Up
          Serial.println(F("Ring up"));
          ring_up();
          keystroke = ' ';
          break;
        }
      case 'R':
      case 'r': {
          // ----- Ring Down
          Serial.println(F("Ring down"));
          ring_down();
          keystroke = ' ';
          break;
        }
      case '%':
      case '5': {
          // ----- Little Up
          Serial.println(F("Little up"));
          little_up();
          keystroke = ' ';
          break;
        }
      case 'T':
      case 't': {
          // ----- Little Down
          Serial.println(F("Little down"));
          little_down();
          keystroke = ' ';
          break;
        }
      case 'X':
      case 'x': {
          // ----- exit
          Serial.println(F("Calibration complete ..."));
          Serial.println(F(" "));
          keystroke = 'X';
          break;
        }
      // ----- default for keystroke
      default: {
          break;
        }
    }
  }
}

// ----------------------------------
// thumb_up()
// ----------------------------------
void thumb_up() {
  for (short i = 0; i < 500; i++) {
    THUMB_step = true;
    THUMB_dir = CW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// thumb_down()
// ----------------------------------
void thumb_down() {
  for (short i = 0; i < 500; i++) {

    THUMB_step = true;
    THUMB_dir = CCW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// index_up()
// ----------------------------------
void index_up() {
  for (short i = 0; i < 500; i++) {
    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = true;
    INDEX_dir = CW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// index_down()
// ----------------------------------
void index_down() {
  for (short i = 0; i < 500; i++) {

    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = true;
    INDEX_dir = CCW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// middle_up()
// ----------------------------------
void middle_up() {
  for (short i = 0; i < 500; i++) {
    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = true;
    MIDDLE_dir = CW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// middle_down()
// ----------------------------------
void middle_down() {
  for (short i = 0; i < 500; i++) {

    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = true;
    MIDDLE_dir = CCW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// ring_up()
// ----------------------------------
void ring_up() {
  for (short i = 0; i < 500; i++) {
    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = true;
    RING_dir = CW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// ring_down()
// ----------------------------------
void ring_down() {
  for (short i = 0; i < 500; i++) {

    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = true;
    RING_dir = CCW;

    LITTLE_step = false;
    LITTLE_dir = CCW;

    step_motors();
  }
}

// ----------------------------------
// little_up()
// ----------------------------------
void little_up() {
  for (short i = 0; i < 500; i++) {
    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = true;
    LITTLE_dir = CW;

    step_motors();
  }
}

// ----------------------------------
// little_down()
// ----------------------------------
void little_down() {
  for (short i = 0; i < 500; i++) {

    THUMB_step = false;
    THUMB_dir = CCW;

    INDEX_step = false;
    INDEX_dir = CCW;

    MIDDLE_step = false;
    MIDDLE_dir = CCW;

    RING_step = false;
    RING_dir = CCW;

    LITTLE_step = true;
    LITTLE_dir = CCW;

    step_motors();
  }
}
